@@ -617,36 +617,6 @@ test_expect_success 'renaming to bogus section is rejected' '
 	test_must_fail git config --rename-section branch.zwei "bogus name"
 '
 
-test_expect_success 'renaming a section with a long line' '
-	{
-		printf "[b]\\n" &&
-		printf "  c = d %1024s [a] e = f\\n" " " &&
-		printf "[a] g = h\\n"
-	} >y &&
-	git config -f y --rename-section a xyz &&
-	test_must_fail git config -f y b.e
-'
-
-test_expect_success 'renaming an embedded section with a long line' '
-	{
-		printf "[b]\\n" &&
-		printf "  c = d %1024s [a] [foo] e = f\\n" " " &&
-		printf "[a] g = h\\n"
-	} >y &&
-	git config -f y --rename-section a xyz &&
-	test_must_fail git config -f y foo.e
-'
-
-test_expect_success 'renaming a section with an overly-long line' '
-	{
-		printf "[b]\\n" &&
-		printf "  c = d %525000s e" " " &&
-		printf "[a] g = h\\n"
-	} >y &&
-	test_must_fail git config -f y --rename-section a xyz 2>err &&
-	grep "refusing to work with overly long line in .y. on line 2" err
-'
-
 cat >> .git/config << EOF
   [branch "zwei"] a = 1 [branch "vier"]
 EOF
@@ -2054,8 +2024,17 @@ test_expect_success '--show-scope with --list' '
 	local	user.override=local
 	local	include.path=../include/relative.include
 	local	user.relative=include
+	local	core.repositoryformatversion=1
+	local	extensions.worktreeconfig=true
+	worktree	user.worktree=true
 	command	user.cmdline=true
 	EOF
+	git worktree add wt1 &&
+	# We need these to test for worktree scope, but outside of this
+	# test, this is just noise
+	test_config core.repositoryformatversion 1 &&
+	test_config extensions.worktreeConfig true &&
+	git config --worktree user.worktree true &&
 	git -c user.cmdline=true config --list --show-scope >output &&
 	test_cmp expect output
 '
@@ -2104,12 +2083,13 @@ test_expect_success '--show-scope with --show-origin' '
 '
 
 test_expect_success 'override global and system config' '
-	test_when_finished rm -f "$HOME"/.config/git &&
-
+	test_when_finished rm -f \"\$HOME\"/.gitconfig &&
 	cat >"$HOME"/.gitconfig <<-EOF &&
 	[home]
 		config = true
 	EOF
+
+	test_when_finished rm -rf \"\$HOME\"/.config/git &&
 	mkdir -p "$HOME"/.config/git &&
 	cat >"$HOME"/.config/git/config <<-EOF &&
 	[xdg]
